/*
**  Copyright (c) 1991 Bolt Beranek and Newman, Inc.
**  All rights reserved.
**
**  Redistribution and use in source and binary forms are permitted
**  provided that: (1) source distributions retain this entire copyright
**  notice and comment, and (2) distributions including binaries display
**  the following acknowledgement:  ``This product includes software
**  developed by Bolt Beranek and Newman, Inc. and CREN/CSNET'' in the
**  documentation or other materials provided with the distribution and in
**  all advertising materials mentioning features or use of this software.
**  Neither the name of Bolt Beranek and Newman nor CREN/CSNET may be used
**  to endorse or promote products derived from this software without
**  specific prior written permission.
**
**  THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
**  WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
**  MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
*/
/*
 * Copyright (c) 1992 Purdue University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Purdue University.  The name of the University may not be used
 * to endorse or promote products derived * from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Note: this copyright applies to portions of this software developed
 * at Purdue beyond the software covered by the original copyright.
 */
#include <stdio.h>
#include <malloc.h>
#include <signal.h>
#include <errno.h>
#include <setjmp.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <sys/ioctl.h>
#include <sys/time.h>
#include <sys/file.h>
#include <sys/fcntl.h>
#include <sys/param.h>
#include <sys/stream.h>
#include <sys/resource.h>
#include <netinet/in.h>
#include <netinet/in_systm.h>
#include <netinet/ip.h>
#include <netinet/tcp.h>
#include <sys/socket.h>
#include <net/if.h>
#include <arpa/inet.h>
#include <stropts.h>

#ifndef SOLARIS_1
#include <sys/dlpi.h>
#endif

#include "args.h"
#include "dpio.h"
#include "dp.h"
#include "dpd.h"
#include "ppp.h"
#include "dpvar.h"

static int	reqtrace;

static jmp_buf	context;
static int	pleaserescan;
static int	pleasequit;
static char	*WHERE = "dpd";

void		dlattachreq	__ARGS((int, u_long));
static void	catchsigterm	__ARGS((int));
static void	catchsighup	__ARGS((int));
static void	catchsigchld	__ARGS((int));
static void	usage		__ARGS(());
void		main		__ARGS((int, char **));
static void	trace_pkt	__ARGS((register REMOTE *,
					register dl_unitdata_ind_t *,
					int, char *, int));
void		mem_report	__ARGS(());

#ifdef SOLARIS_1
#define DEVDIAL	"/dev/dp0"
#endif

/*ARGUSED*/
static void
catchsigterm(sig)
int sig;
{
    pleasequit++;
    longjmp(context, 1);
}


/*ARGUSED*/
static void
catchsighup(sig)
int sig;
{
    pleaserescan++;
    (void)signal(SIGHUP, catchsighup);
    longjmp(context, 1);
}


/*ARGUSED*/
static void
catchsigchld(sig)
int sig;
{
    int w;
    int stat;

    errno = 0;
    while ((w = waitpid((pid_t)-1, &stat, WNOHANG)) > 0) {
	if (WIFEXITED(stat) && WEXITSTATUS(stat) != 0)
	    d_log(DLOG_GENERAL, WHERE, "child %d exited with status %d\n",
		  w, WEXITSTATUS(stat));
	else if (WIFSIGNALED(stat))
	    d_log(DLOG_GENERAL, WHERE, "child %d terminated due to signal %d\n",
		  w, WTERMSIG(stat));
    }
    (void)signal(SIGCHLD, catchsigchld);
}



static void
usage()
{
    (void)fprintf(stderr, "usage: %s [options]\n", progname);
    d_log(DLOG_GENERAL, WHERE, "Usage error");
    exit(10);
}


void
main(argc, argv)
int argc;
char **argv;
{
    int			fd;
    int			f;
    int			nofork;
    int			i;
    char		dlu[128];
    char		pkt[PPP_MTU];
    REMOTE		*rp;
    char		*configfile;
    struct strbuf ctlb, datab;
    int 		npfds;
    struct pollfd	*pfds,
			*pfd;
    int			sf;
    struct rlimit	rl;
    ulong		sap;
#ifdef SOLARIS_1
    struct dp_req	*dpreq;
    dl_unitdata_ind_t	*dlup;
    phdr_t		phdr;
    char		ifname[IFNAMSIZ];
#endif

    /* Set defaults. */
    setprogname(argv[0]);
    WHERE = progname;
    nofork = 0;
    configfile = DP_CONFIG;

    /* Parse flags. */
    while ((i = getopt(argc, argv, "c:d:Fft")) != EOF)
	switch (i) {
	default:
	    usage();
	    /* NOTREACHED */
	case 'c':
	    configfile = optarg;
	    break;
	case 'd':
	    log_level = atoi(optarg);
	    break;
	case 'f':
	    nofork = 1;
	    break;
	case 'F':
	    nofork = 2;
	    break;
	case 't':
	    reqtrace = 1;
	    break;
	}
    argc -= optind;
    argv += optind;
    if (argc)
	usage();

    /* Fork us off */
    if (nofork < 1) {
	if ((i = fork()) < 0) {
	    d_log(DLOG_GENERAL, WHERE, "Can't fork, %m");
	    exit(3);
	}
	if (i > 0)
	    exit(0);

    }

    /*
     * Sanitize our process state.
     */
    if (nofork < 1) {
	if (getrlimit(RLIMIT_NOFILE, &rl) == 0) {
	    for (f = rl.rlim_cur - 1 ; f >= 0 ; f--)
		(void)close(f);
	}
	(void)open("/", O_RDONLY);
	(void)dup2(0, 1);
	(void)dup2(0, 2);
	(void)setsid();
    }
    else
	if (getrlimit(RLIMIT_NOFILE, &rl) == 0) {
	    for (f = rl.rlim_cur ; f > 2 ; f--)
		(void)close(f);
	}

    /*
     * Figure out where everything is located.
     */
    init_pathconf();

    /*
     * Read network and modem configuration.
     */
    if (!readconfig(configfile))
	exit(4);

    if (!setfilelimit())
	exit(5);

#ifdef SOLARIS_1

    /* Open the request device. */
    if ((f = open(DEVDIAL, O_RDONLY)) < 0) {
	d_log(DLOG_GENERAL, WHERE, "Can't open \"%s\", %m", DEVDIAL);
	exit(1);
    }
#ifdef	FIOCLEX
    if (ioctl(f, FIOCLEX, (caddr_t)NULL) < 0)
	d_log(DLOG_INFO, WHERE, "Can't do FIOCLEX on \"%s\", %m", DEVDIAL);
#endif	/* FIOCLEX */

#else /* SOLARIS_2 */

    /*
     * Open the DP devices.
     */
    if (!(npfds = opendpdevs(&pfds)))
	exit(6);

#endif /* SOLARIS_1 or SOLARIS_2 */

    /* Tell the log file where are here */
    d_log(DLOG_GENERAL, WHERE, "%s started", dp_release());
    record_pid(progname, getpid(), -1);

    /*
     * Initialize timeouts.
     */
    if (!set_if_timeouts())
	exit(7);

    /* Process requests */
    (void)signal(SIGHUP, catchsighup);
    (void)signal(SIGTERM, catchsigterm);
    (void)signal(SIGCHLD, catchsigchld);
    for ( ; ; ) {
	(void)setjmp(context);
	if (pleasequit)
	    break;
	if (pleaserescan) {
	    d_log(DLOG_GENERAL, WHERE, "Rescanning config files");
#ifdef	MDEBUG
	    mem_report();
#endif
#ifndef SOLARIS_1
	    closedpdevs();
#endif
	    if (!readconfig(configfile)) {
		unlock_pid();
		exit(8);
	    }
#ifndef SOLARIS_1
	    if (!(npfds = opendpdevs(&pfds))) {
		unlock_pid();
		exit(9);
	    }
#endif
	    (void)signal(SIGHUP, catchsighup);
	    pleaserescan = 0;
	    continue;
	}

	/*
	 * get next dial request ...
	 */

#ifdef SOLARIS_1

	ctlb.maxlen = -1;
	ctlb.buf = (char *)0;
	datab.maxlen = sizeof(struct dp_req);
	datab.buf = pkt;
	sf = 0;
	if (i = getmsg(f, &ctlb, &datab, &sf)) {
	    if (errno != EINTR)
		d_log(DLOG_GENERAL, WHERE,
		    "Bad getmsg (returned %d) from \"%s\", %m",
		    i, DEVDIAL);
	    continue;
	}
	dpreq = (struct dp_req *)pkt;
	if (datab.len < sizeof(dp_reqinfo) + DR_MINHDRLEN ||
	    datab.len > sizeof(dp_reqinfo) + DR_MAXHDRLEN ||
	    datab.len != sizeof(dp_reqinfo) + dpreq->dr_hdrlen) {
		d_log(DLOG_GENERAL, WHERE,
		    "Request packet length (%d) out of range", dpreq->dr_hdrlen);
	    continue;
	}
	if (dpreq->dr_hdrlen < DR_MINHDRLEN || dpreq->dr_hdrlen > DR_MAXHDRLEN) {
		d_log(DLOG_GENERAL, WHERE,
		    "Request header length (%d) out of range", dpreq->dr_hdrlen);
	    continue;
	}

	/* Find config data for the address. */
	(void)sprintf(ifname, "%s%d", dpreq->dr_ifname, dpreq->dr_ifunit);
	if ((rp = findconfig(ifname)) == NULL) {
	    d_log(DLOG_GENERAL, WHERE, "Address %s not configured",
		inet_ntoa(dpreq->dr_sin.sin_addr));
	    continue;
	}

	/*
	 * now fake up a DLPI unitdata indication
	 */

	dlup = (dl_unitdata_ind_t *)dlu;
	dlup->dl_primitive = DL_UNITDATA_IND;
	dlup->dl_dest_addr_length = DP_DLSAP_SIZE;
	dlup->dl_dest_addr_offset = sizeof (dl_unitdata_ind_t);
	dlup->dl_src_addr_length = DP_DLSAP_SIZE;
	dlup->dl_src_addr_offset = sizeof (dl_unitdata_ind_t) + DP_DLSAP_SIZE;
	dlup->dl_group_address = 0;
	phdr.ph_address = PPP_ALLSTATIONS;
	phdr.ph_control = PPP_UI;
	phdr.ph_protocol = PPP_IP;
	memcpy(dlu+dlup->dl_dest_addr_offset, (caddr_t *)&phdr, DP_DLSAP_SIZE);
	memcpy(dlu+dlup->dl_src_addr_offset, (caddr_t *)&phdr, DP_DLSAP_SIZE);

	ctlb.maxlen = sizeof(dlu);
	ctlb.buf = dlu;
	ctlb.len = sizeof (dl_unitdata_ind_t) + 2*(DP_DLSAP_SIZE);

	datab.buf += sizeof(dp_reqinfo);	/* skip to IP header */
	datab.len -= sizeof(dp_reqinfo);

#else /* SOLARIS_2 */

	for (fd = 0, pfd = pfds ; fd < npfds ; fd++, pfd++)
	    pfd->revents = 0;
	if (poll(pfds, npfds, -1) < 0) {
	    if (errno == EINTR)
		continue;
	    d_log(DLOG_GENERAL, WHERE,
		  "Bad poll (returned %d) from \"%s\", %m", i, DEVDP);
	    unlock_pid();
	    exit(10);
	}
	for (fd = 0, pfd = pfds ; fd < npfds ; fd++, pfd++) {
	    if (!pfd->revents)
		continue;
	    /*
	     * Find config data for this PPA.
	     */
	    if ((rp = findfdconfig(pfd->fd)) == NULL) {
		d_log(DLOG_GENERAL, WHERE, "No REMOTE for fd %d ?!?", pfd->fd);
		pfd->events = 0;
		continue;
	    }
	    
	    if (pfd->revents != POLLIN) {
		d_log(DLOG_GENERAL, WHERE,
		    "Bad poll return events 0x%x from \"%s\", %m",
		    i, rp->Device);
		pfd->events = 0;
		continue;
	    }

	    ctlb.maxlen = sizeof(dlu);
	    ctlb.buf = dlu;
	    datab.maxlen = sizeof(pkt);
	    datab.buf = pkt;
	    sf = 0;
	    if (i = getmsg(pfd->fd, &ctlb, &datab, &sf)) {
		if (errno != EINTR)
		    d_log(DLOG_GENERAL, WHERE,
			"Bad getmsg (returned %d) from \"%s\", %m",
			i, DEVDP);
		continue;
	    }

#endif /* SOLARIS_1 or SOLARIS_2 */

	    if (reqtrace)
		trace_pkt(rp,
			  (dl_unitdata_ind_t *)ctlb.buf, ctlb.len,
			  datab.buf, datab.len);

	    /* Is this packet is authorized? */
	    if (!authorized(rp->Access, (dl_unitdata_ind_t *)ctlb.buf, ctlb.len,
			    datab.buf, datab.len)) {
		failcall(rp, (MODEM *)0, DP_FAILURE, "authorization failure");
		continue;
	    }

	    /* Make the call. */
	    if (nofork >= 2) {
		if (rp->LogLevel > log_level)
		    log_level = rp->LogLevel;
		req_log((dl_unitdata_ind_t *)ctlb.buf, datab.buf);
		makecall(rp, (dl_unitdata_ind_t *)ctlb.buf, datab.buf);
		continue;
	    }
	    if ((i = fork()) == 0) {
		(void)setpgrp(0, getpgrp(0));
		(void)close(f);
		if (rp->LogLevel > log_level)
		    log_level = rp->LogLevel;
		req_log((dl_unitdata_ind_t *)ctlb.buf, datab.buf);
		makecall(rp, (dl_unitdata_ind_t *)ctlb.buf, datab.buf);
		exit(0);
	    }
	    if (i < 0)
		d_log(DLOG_GENERAL, WHERE, "Can't fork dialing process, %m");
#ifndef SOLARIS_1
	}
#endif
    }

    d_log(DLOG_GENERAL, WHERE, "%s exiting", dp_release());
    unlock_pid();
    exit(0);
    /* NOTREACHED */
}

static void
trace_pkt(rp, dlu, dlu_len, pkt, pkt_len)
register REMOTE *rp;
register dl_unitdata_ind_t *dlu;
int dlu_len;
char *pkt;
int pkt_len;
{
    int fd;
    pkt_trace_t pt;
    char *tbuf, *t;
    char *tf = expand_dir_file("$DPLOG_DIR", TRACE_FILE);
    if ((fd = open(tf, O_WRONLY|O_APPEND|O_CREAT, 0644)) >= 0) {
	pt.time = time((time_t *)0);
	pt.ppa = rp->Unit;
	pt.dlu_len = dlu_len;
	pt.pkt_len = pkt_len;
	t = tbuf = malloc(sizeof(pt) + dlu_len + pkt_len);
	(void)memcpy(t, (char *)&pt, sizeof(pt));
	t += sizeof(pt);
	(void)memcpy(t, (char *)dlu, dlu_len);
	t += dlu_len;
	(void)memcpy(t, (char *)pkt, pkt_len);
	(void)write(fd, tbuf, sizeof(pt) + dlu_len + pkt_len);
	(void)close(fd);
	free(tbuf);
    }
    (void)free(tf);
}

#ifdef	MDEBUG
#include <malloc.h>

void
mem_report()
{
    struct mallinfo ma;
    ma = mallinfo();
    d_log(DLOG_GENERAL, WHERE,
	  "arena = %5d, usmblks = %5d, uordblks = %5d",
	  ma.arena, ma.usmblks, ma.uordblks);
}
#endif
