/*
 * ppp.c - Point-to-Point Protocol.
 *
 * Copyright (c) 1989 Carnegie Mellon University.
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Carnegie Mellon University.  The name of the
 * University may not be used to endorse or promote products derived
 * from this software without specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
/*
 * Copyright (c) 1992 Purdue University
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms are permitted
 * provided that the above copyright notice and this paragraph are
 * duplicated in all such forms and that any documentation,
 * advertising materials, and other materials related to such
 * distribution and use acknowledge that the software was developed
 * by Purdue University.  The name of the University may not be used
 * to endorse or promote products derived * from this software without
 * specific prior written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED
 * WARRANTIES OF MERCHANTIBILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 * Note: this copyright applies to portions of this software developed
 * at Purdue beyond the software covered by the original copyright.
 */

/*
 * TODO:
 * Don't prompt for passwd on /dev/tty (go into server mode).
 * Might want to choose IP addresses from config file based on user id
 * if the server, or based on remote host if tip'd client.
 * Way to set asyncmap.
 * Wait for tty queue to empty before exiting in quit.
 */

/*
 * There are three scenarios:
 * 1.  ppp used as daemon started from /etc/rc or perhaps /etc/ttys.
 *	a.  server
 *	b.  authentication necessary
 *	c.  want to use constant local ip addr
 *	d.  want to use constant remote ip addr, constant ip addr based on
 *	    authenticated user, or request ip addr
 * 2.  ppp used on /dev/tty after remote login.
 *	a.  server
 *	b.  no authentication necessary or allowed
 *	c.  want to use constant local ip addr
 *	d.  want to use constant remote ip addr, constant ip addr based on
 *	    authenticated user, or request ip addr
 * 3.  ppp used on line after tip'ing out.
 *	a.  client
 *	b.  remote end may request authentication
 *	c.  want to use constant local ip addr or request ip addr
 *	d.  want to use constant remote ip addr based on tip'd host, or
 *	    request remote ip addr
 */

#include <stdio.h>
#include <signal.h>
#include <errno.h>
#include <fcntl.h>
#include <pwd.h>
#include <syslog.h>
#include <netdb.h>
#include <utmp.h>

#include <sys/param.h>
#include <sys/types.h>
#include <sys/file.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <sys/sockio.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <sys/time.h>
#include "callout.h"

#ifndef SOLARIS_1
#include <sys/dlpi.h>
#endif

#include <net/if.h>
#include <net/route.h>
#include <sys/stream.h>
#include <sys/stropts.h>
#include <sys/termios.h>
#include <sys/conf.h>
#include <poll.h>

#define	DEVNAME_SIZE	128	/* Buffer size for /dev filenames */

#include <malloc.h>
#include <string.h>


#include "ppp.h"
#include "magic.h"
#include "fsm.h"
#include "lcp.h"
#include "ipcp.h"
#include "vjc.h"
#include "upap.h"
#include "chap.h"
#include "pathnames.h"
#include "patchlevel.h"

#include "dpio.h"
#include "slip.h"

#define PPPHOSTS	"/etc/hosts.ppp"

#ifndef TRUE
#define TRUE (1)
#endif /*TRUE*/

#ifndef FALSE
#define FALSE (0)
#endif /*FALSE*/

#ifdef PIDPATH
static char *pidpath = PIDPATH;	/* filename in which pid will be */
				/* stored */
#else
static char *pidpath = _PATH_PIDFILE;
#endif /* PIDFILE */

static char uinfopath[DEVNAME_SIZE];

/* global vars */

int phase;

int debug = 0;		        /* Debug flag */
char ifname[IFNAMSIZ];		/* Interface name */

int fd;				/* Async Device file descriptor */
#ifndef SOLARIS_1
int dpfd = -1;			/* Network device file descriptor */
int dpmuxid = -1;		/* muxid for Async attachment to dp device */
int dpipfd = -1;		/* Network device file descriptor for IP */
int ipfd = -1;			/* IP device file descriptor */
int ipmuxid = -1;		/* muxid for attachment to IP device */
#endif

char hostname[MAX_HOSTNAME_LEN]; /* hostname */
u_char hostname_len;		/* hostname length */
u_char outpacket_buf[MTU+DLLHEADERLEN]; /* buffer for outgoing packet */

char devname[DEVNAME_SIZE] = "/dev/tty";	/* Device name */
char passwd[MAXSECRETLEN];		/* passwd */
char user[MAXNAMELEN];		/* username for PAP */

char our_name[MAXNAMELEN];
char remote_name[MAXNAMELEN];
int auth_required = 0;		/* require peer to authenticate */
int uselogin = 0;		/* check PAP info against /etc/passwd */
int usehostname = 0;

/* local vars */
static char *progname;			/* Name of this program */
static int ifunit = -1;			/* Interface unit number */
static char inspeed = 0;	        /* Input/Output speed */
static char outspeed = 0;	        /* Input/Output speed */
#ifdef SOLARIS_1
static int initfdflags;		/* Initial file descriptor flags */
#endif
int s;				/* Socket file descriptor */
static pid_t	pid;			/* Our pid */
#ifdef sun
static pid_t	pgrpid;			/* Process Group ID */
#endif

int dialup = 0;			/* Dialup mode - set by "dialup" argument

				   Historically, this was an attempt
				   to distinguish at runtime between
				   the "standard" pppd and dialup-ppp (dp).
				   I very much doubt if running with
				   dialup = 0 would do anything remotely
				   useful now.  tkr 09feb95 

				 */
  
int dynamic = 0;		/* Allow dynamic addresses */
int syncmode = FALSE;		/* Synchronous HDLC mode */
static int slip = 0;		/* Use SLIP framing */

static int default_device = TRUE; /* use default device (stdin/out) */

static u_char inpacket_buf[MTU+DLLHEADERLEN]; /* buffer for incoming packet */

static struct in_addr netmask;	/* netmask to use on ppp interface */
#if 0
static char pidfilename[DEVNAME_SIZE];
#endif

static void hup __ARGS((int));
static void intr __ARGS((int));
static void term __ARGS((int));
static void alrm __ARGS((int));
static void io __ARGS((int));
static void incdebug __ARGS((int));
static void nodebug __ARGS((int));
static void getuserpasswd __ARGS((void));

static int setdebug __ARGS((int *, char ***));
static int setdialup __ARGS((int *, char ***));
static int setdynamic __ARGS((int *, char ***));
static int setunit __ARGS((int *, char ***));
static int setpassive __ARGS((int *, char ***));
static int setsilent __ARGS((int *, char ***));
static int noopt __ARGS((int *, char ***));
static int setnovj __ARGS((int *, char ***));
static int setnovjccomp __ARGS((int *, char ***));
static int setvjslots __ARGS((int *, char ***));
static int noupap __ARGS((int *, char ***));
static int requpap __ARGS((int *, char ***));
static int nochap __ARGS((int *, char ***));
static int reqchap __ARGS((int *, char ***));
static int setspeed __ARGS((int *, char ***));
static int noaccomp __ARGS((int *, char ***));
static int noasyncmap __ARGS((int *, char ***));
static int noipaddr __ARGS((int *, char ***));
static int nomagicnumber __ARGS((int *, char ***));
static int setasyncmap __ARGS((int *, char ***));
static int setvjmode __ARGS((int *, char ***));
static int setmru __ARGS((int *, char ***));
static int nomru __ARGS((int *, char ***));
static int nopcomp __ARGS((int *, char ***));
static int setdomain __ARGS((int *, char ***));
static int setnetmask __ARGS((int *, char ***));
static int setname __ARGS((int *, char ***));
static int setusername __ARGS((int *, char ***));
static int setremote __ARGS((int *, char ***));
static int setauth __ARGS((int *, char ***));
static int setdologin __ARGS((int *, char ***));
static int setusehostname __ARGS((int *, char ***));
static int setpaptimeout __ARGS((int *, char ***));
static int setpapreqs __ARGS((int *, char ***));
static int setchaptimeout __ARGS((int *, char ***));
static int setchapchal __ARGS((int *, char ***));
static int setchapintv __ARGS((int *, char ***));
static int setsync __ARGS((int *, char ***));
static int setslip __ARGS((int *, char ***));
#ifdef SOLARIS_1
static void cleanup __ARGS((int, caddr_t));
#else
static void cleanup __ARGS((void));
#endif
static void str_restore __ARGS((void));
extern	char	*ttyname __ARGS((int));
#define	MAXMODULES	10		/* max number of module names that we can save */
static struct	modlist {
  char	modname[FMNAMESZ+1];
} str_modules[MAXMODULES];
static int	str_module_count = 0;
static int	str_push_count = 0;
static struct termios saved_tty_parameters;

char *popstoplist[] = {
    "sockmod",
    "ptem",
    "telnet",
    "tirdwr",
    "timod",
    "advmod",
    (char *)0
};

/*
 * Valid arguments.
 */
static struct cmd {
  char *cmd_name;
  int (*cmd_func)();
} cmds[] = {
  "-all", noopt,		/* Don't request/allow any options */
  "-ac", noaccomp,		/* Disable Address/Control compress */
  "-am", noasyncmap,		/* Disable asyncmap negotiation */
  "-as", setasyncmap,		/* set the desired async map */
  "-d", setdebug,		/* Increase debugging level */
  "debug", setdebug,		/* Increase debugging level */
  "dialup", setdialup,		/* Do special stuff for dialup */
  "-ip", noipaddr,		/* Disable IP address negotiation */
  "-mn", nomagicnumber,	        /* Disable magic number negotiation */
  "-mru", nomru,		/* Disable mru negotiation */
  "mru", setmru,		/* Set MRU value for negotiation */
  "-p", setpassive,		/* Set passive mode */
  "-s", setsilent,		/* Set silent mode */
  "-pc", nopcomp,		/* Disable protocol field compress */
  "passive", setpassive,	/* Set passive mode */
  "silent", setsilent,		/* Set silent mode */
  "+pap", requpap,		/* Require UPAP authentication */
  "-pap", noupap,		/* Don't allow UPAP authentication */
  "+chap", reqchap,		/* Require CHAP authentication */
  "-chap", nochap,		/* Don't allow CHAP authentication */
  "vjmode", setvjmode,	        /* set VJ compression mode */
  "-vj", setnovj,		/* disable VJ compression */
  "-vjccomp", setnovjccomp,	/* disable VJ connection-ID compression */
  "vj-max-slots", setvjslots,	/* Set maximum VJ header slots */
  "sync", setsync,		/* operate in synchronous HDLC mode */
  "domain", setdomain,	        /* Add given domain name to */
				/* hostname*/
  "netmask", setnetmask,        /* set netmask */
  "unit", setunit,		/* Set Unit request */
  "slip", setslip,		/* Do SLIP framing */
  "dynamic", setdynamic,	/* Allow Dynamic addresses */
  "name", setname,		/* Set local name for authentication */
  "user", setusername,		/* Set username for PAP auth with peer */
  "usehostname", setusehostname,	/* Must use hostname for auth. */
  "remotename", setremote,	/* Set remote name for authentication */
  "auth", setauth,		/* Require authentication from peer */
  "login", setdologin,		/* Use system password database for UPAP */
  "pap-restart", setpaptimeout,	/* Set timeout for UPAP */
  "pap-max-authreq", setpapreqs,	/* Set max #xmits for auth-reqs */
  "chap-restart", setchaptimeout,	/* Set timeout for CHAP */
  "chap-max-challenge", setchapchal,	/* Set max #xmits for challenge */
  "chap-interval", setchapintv,	/* Set interval for rechallenge */
#if 0
  "local", setlocalipaddr,
  "remote", setremoteipaddr,
  "test", enabletestline,
#endif
  NULL
  };

void log_packet __ARGS((u_char *, int, char *));
void format_packet __ARGS((u_char *, int,
			   void (*) (void *, char *, ...), void *));
void pr_log __ARGS((void *, char *, ...));

/*
 * PPP Data Link Layer "protocol" table.
 * One entry per supported protocol.
 */
static struct protent {
    u_short protocol;
    void (*init)();
    void (*input)();
    void (*protrej)();
    int  (*printpkt)();
    char *name;
} prottbl[] = {
    { LCP, lcp_init, lcp_input, lcp_protrej, lcp_printpkt, "LCP" },
    { IPCP, ipcp_init, ipcp_input, ipcp_protrej, ipcp_printpkt, "IPCP" },
    { UPAP, upap_init, upap_input, upap_protrej, upap_printpkt, "PAP" },
    { CHAP, ChapInit, ChapInput, ChapProtocolReject, ChapPrintPkt, "CHAP" },
};

#define N_PROTO		(sizeof(prottbl) / sizeof(prottbl[0]))

char *usage = "PPP %s patch level %d\n\
Usage: %s [ arguments ], where arguments are:\n\
	-all		Don't request/allow any options\n\
	-ac		Disable Address/Control compression\n\
	-am		Disable asyncmap negotiation\n\
	-as <n>		Set the desired async map to hex <n>\n\
	-d		Increase debugging level\n\
	debug		Increase debugging level\n\
	dialup		Do special stuff for dialup\n\
	-ip		Disable IP address negotiation\n\
	-mn		Disable magic number negotiation\n\
	-mru		Disable mru negotiation\n\
	mru <n>		Set MRU value to <n> for negotiation\n\
	-p		Set passive mode\n\
	-s		Set silent mode\n\
	-pc		Disable protocol field compression\n\
	passive		Set passive mode\n\
	silent		Set silent mode\n\
	+pap		Require UPAP authentication\n\
	-pap		Don't allow UPAP authentication\n\
	+chap		Require CHAP authentication\n\
	-chap		Don't allow CHAP authentication\n\
	vjmode <m>      VJ compression mode {rfc1172, pre1332, rfc1332}\n\
        -vj             disable VJ compression\n\
	-vjcomp		disable VJ connection-ID compression\n\
	vj-max-slots	Set maximum VJ header slots\n\
	sync		Operate in synchronous HDLC mode\n\
        domain <d>      Append domain name <d> to hostname for authentication\n\
	netmask		set netmask\n\
	unit		Set Unit request\n\
	slip		Do SLIP framing\n\
	dynamic		Allow Dynamic addresses\n\
	name		Set local name for authentication\n\
	user		Set username for PAP auth with peer\n\
	usehostname	Must use hostname for auth.\n\
	remotename	Set remote name for authentication\n\
	auth		Require authentication from peer\n\
	login		Use system password database for UPAP\n\
	pap-restart	Set timeout for UPAP\n\
	pap-max-authreq	Set max #xmits for auth-reqs\n\
	chap-restart	Set timeout for CHAP\n\
	chap-max-challenge Set max #xmits for challenge\n\
	chap-interval	Set interval for rechallenge\n\
	<device>	Communicate over the named device\n\
	<speed>		Set the baud rate to <speed>\n\
	<loc>:<rem>	Set the local and/or remote interface IP\n\
			addresses.  Either one may be omitted.\n";

/* 
 * this module will attempt to reconstruct the stream with the
 * previously popped modules
 */

/*ARGSUSED*/
static void
str_restore()
{
  /*EMPTY*/ 

#ifndef SOLARIS_1
  errno = 0;
  if(dpfd >= 0) {
    if((dpmuxid != -1) && (ioctl(dpfd, I_UNLINK, dpmuxid) < 0)) {
      MAINDEBUG((LOG_DEBUG, "str_restore: couldn't unlink from dp mux: %m"))
    }
    dpmuxid = -1;
    close(dpfd);
    dpfd = -1;
  }
#endif

    /* pop any we pushed */
    for ( ;
	 (str_push_count > 0) && (ioctl (fd, I_POP, 0) >= 0) ;
	 str_push_count--)
	;

    for ( ; str_module_count > 0 ; str_module_count--) {
	if (ioctl (fd, I_PUSH, str_modules[str_module_count - 1].modname)) {
	    MAINDEBUG ((LOG_ERR, "str_restore: couldn't push module %s: %m",
		       str_modules[str_module_count - 1].modname))
	}
	else {
	    MAINDEBUG ((LOG_INFO, "str_restore: pushed module %s",
		       str_modules[str_module_count - 1].modname))
	}
    }

#ifndef	SOLARIS_1
    if (ipfd >= 0) {
	if (ipmuxid != -1) {
	    if (ioctl(ipfd, I_UNLINK, ipmuxid) < 0) {
		MAINDEBUG((LOG_ERR,
			  "str_restore: couldn't unlink from ip mux: %m"))
	    }
	    ipmuxid = -1;
	}
	close(ipfd);
	ipfd = -1;
    }

    if (dpipfd >= 0) {
	ioctl(dpipfd, I_POP, 0); /* pop IP module */
	if (strioctl(dpipfd, DPIOCIFDETACH, -1, sizeof(int), (caddr_t)&ifunit)
	    < 0) {
	    MAINDEBUG((LOG_DEBUG, "str_restore: DPIOCIFDETACH: %m"))
	}
	close(dpipfd);
	dpipfd = -1;
    }
#endif	/* not SOLARIS_1 */

    if (!syncmode)
	ioctl (fd, TCSETS, &saved_tty_parameters);
}

main(argc, argv)
int argc;
char *argv[];
{
  int i;
  sigset_t mask;
  struct termios	tios;
  struct sigaction sa;
  struct cmd *cmdp;
#ifndef SOLARIS_1
  char buf[DP_MAXDLBUF];
  ulong sap;
#endif
  char *p;

#if 0
  FILE *pidfile;
#endif

  /*
   * Initialize syslog system and magic number package.
   */
  openlog("ppp", LOG_PID | LOG_NDELAY, LOG_LOCAL2);
  setlogmask(LOG_UPTO(LOG_WARNING));
  
  p = ttyname(fileno(stdin));
  if (p)
    strcpy(devname, ttyname(fileno(stdin)));

  magic_init();

  if (gethostname(hostname, MAX_HOSTNAME_LEN) < 0 ) {
    syslog(LOG_ERR, "couldn't get hostname: %m");
    exit(1);
  }

  /*
   * Initialize to the standard option set and then parse the command
   * line arguments.
   */
  for (i = 0; i < N_PROTO; i++)
    (*prottbl[i].init)(0);
  
  if ((progname = strrchr (*argv, '/')))
    progname++;
  else
    progname = *argv;

  if (strstr (progname, "slip"))
    setslip(NULL, NULL);

  for (argc--, argv++; argc; ) {
    /*
     * First see if it's a command.
     */
    for (cmdp = cmds; cmdp->cmd_name; cmdp++)
      if (!strcmp(*argv, cmdp->cmd_name) &&
	  (*cmdp->cmd_func)(&argc, &argv))
	break;
    
    /*
     * Maybe a tty name, speed or IP address?
     */
    if (cmdp->cmd_name == NULL &&
	!setdevname(&argc, &argv) &&
	!setspeed(&argc, &argv) &&
	!setipaddr(&argc, &argv)) {
      if (argc) {
	  fprintf(stderr, "??? %s\n", *argv);
	  syslog(LOG_ERR, "bad PPP_ARG: %s", *argv);
      }
      fprintf(stderr, usage, VERSION, PATCHLEVEL, progname);
      exit(1);
    }
  }
  
  syslog(LOG_INFO, "Starting ppp daemon version %s patchlevel %d",
         VERSION, PATCHLEVEL); 

  check_auth_options();

  /*
   * Initialize state.
   */
#ifndef	SOLARIS_1
  if ((dpipfd = open(DEVDP, O_RDWR)) < 0) {
    syslog(LOG_ERR, "Can't open \"%s\", %m", DEVDP);
    exit(1);
  }

  if(!dialup && (ifunit == -1)) {
    for (ifunit=0; ifunit < 255; ifunit++) {
      if (strioctl(dpipfd, DPIOCIFATTACH, -1, sizeof(int), (caddr_t)&ifunit) < 0) {
	if (errno != EEXIST) {
	  syslog(LOG_ERR, "DPIOCIFATTACH ifunit %d: %m", ifunit);
	  exit(1);
	}
      } else {
	if ((ipfd = open("/dev/ip", O_RDWR)) < 0) {
	  syslog(LOG_ERR, "open(\"/dev/ip\"): %m");
	  exit(1);
	}

	if (ioctl(dpipfd, I_PUSH, "ip") < 0) {
	  syslog(LOG_ERR, "ioctl(I_PUSH, \"ip\"): %m");
	  exit(1);
	}

	if (ioctl(dpipfd, IF_UNITSEL, (caddr_t)&ifunit) < 0) {
	  syslog(LOG_ERR, "ioctl(IF_UNITSEL): %m");
	  exit(1);
	}

	if (strioctl(dpipfd, DPIOCSIFP2PFL, -1, sizeof(int), (caddr_t)&ifunit) < 0) {
	  syslog(LOG_ERR, "ioctl(DPIOCSIFP2PFL): %m");
	  exit(1);
	}

	if ((ipmuxid = ioctl(ipfd, I_LINK, dpipfd)) < 0) {
	  syslog(LOG_ERR, "ioctl(I_LINK): %m");
	  exit(1);
	}
	break;
      }
      if(ifunit >= 255) {
	syslog(LOG_ERR, "no free units");
	exit(1);
      }
    }
  }
#endif	/* not SOLARIS_1 */


  if ((fd = open(devname, O_RDWR /*| O_NDELAY*/)) < 0) {
    syslog(LOG_ERR, "open(%s): %m", devname);
    exit(1);
  }

  if (!syncmode && (ioctl (fd, TCGETS, (caddr_t) & saved_tty_parameters) < 0)) {
      syslog (LOG_ERR, "ioctl(TCGETS): %m");
      /*exit(1); not fatal*/
  }


  if ((s = socket(AF_INET, SOCK_DGRAM, 0)) < 0) {
    syslog(LOG_ERR, "socket : %m");
    exit(1);
  }
  
  /* if we exit, then try and restore the stream */ 
#ifdef SOLARIS_1
  on_exit(cleanup, NULL);
#else
  atexit(cleanup);
#endif
  
  /* go through and save the name of all the modules, then pop em */
  while (1)
    {
      if (!ioctl (fd, I_LOOK, str_modules[str_module_count].modname))
       {
	 char **poplistptr;

	 for (poplistptr = popstoplist; *poplistptr && strcmp (str_modules[str_module_count].modname, *poplistptr); poplistptr++);
	 if (*poplistptr)
	   break;
	 MAINDEBUG ((LOG_DEBUG, "popped stream module : %s",
		     str_modules[str_module_count].modname))
       }
      if (!ioctl (fd, I_POP, 0))
       str_module_count++;
      else
       break;
    }

	if (!syncmode) {
		if (ioctl(fd, TCGETS, (caddr_t)&tios) == -1) {
			syslog(LOG_ERR, "ioctl(TCGETS): %m");
			/* exit(1); not fatal */
		}
		tios.c_cflag = CRTSCTS |
				(inspeed  ? inspeed : (tios.c_cflag & CBAUD));
		tios.c_cflag |= CS8|CREAD|HUPCL;
		tios.c_iflag = IGNBRK;
		tios.c_oflag = 0;
		tios.c_lflag = 0;

		if(ioctl(fd, TCSETS, (caddr_t) &tios) == -1) {
			syslog(LOG_ERR, "ioctl(TCSETS): %m");
			exit(1);
		}
	}

#ifdef SOLARIS_1

  /* now push the async/fcs module */
  if(ioctl(fd, I_PUSH, syncmode ? "pppsync" : "pppasync") < 0) {
    syslog(LOG_ERR, "ioctl(I_PUSH, \"ppp%ssync\"): %m", syncmode ? "" : "a");
    exit(1);
  }
  str_push_count++;
  if (ioctl(fd, I_PUSH, "vjc") < 0) {
    syslog(LOG_ERR, "ioctl(I_PUSH, \"vjc\"): %m");
    exit(1);
  }
  str_push_count++;
  /* finally, push the dpif module that actually
   * handles the network interface
   */
  if(ioctl(fd, I_PUSH, "dpif") < 0) {
    syslog(LOG_ERR, "ioctl(I_PUSH, \"dpif\"): %m");
    exit(1);
  }
  str_push_count++;
  if (ifunit >= 0) {
  /*
   * Select the appropriate network interface before doing
   * anything else with the "dpif" module.
   */
    if(ioctl(fd, DPIOCSIFUNIT, (caddr_t)&ifunit) < 0) {
      syslog(LOG_ERR, "ioctl(DPIOCSIFUNIT): %m");
      exit(1);
    }
  }
  if(ioctl(fd, I_SETSIG, S_INPUT) < 0) {
    syslog(LOG_ERR, "ioctl(I_SETSIG, S_INPUT): %m");
    exit(1);
  }
  /* read mode, message non-discard mode */
  if(ioctl(fd, I_SRDOPT, RMSGN) < 0) {
    syslog(LOG_ERR, "ioctl(I_SRDOPT, RMSGN): %m");
    exit(1);
  }
  /* Flush any waiting messages, or we'll never get SIGPOLL */
  if(ioctl(fd, I_FLUSH, FLUSHRW) < 0) {
    syslog(LOG_ERR, "ioctl(I_FLUSH, FLUSHRW): %m");
    exit(1);
  }
  /*
   * Find out which interface we were given.
   */
  /* in streams, dp_if handles this ioctl */
  if (ioctl(fd, DPIOCGIFUNIT, &ifunit) < 0) {
    syslog(LOG_ERR, "ioctl(DPIOCGIFUNIT): %m");
    exit(1);
  }
  sprintf(ifname, dialup ? "dp%d" : "ppp%d", ifunit);
  syslog(LOG_NOTICE, "Using unit %s", ifname);
  pid = getpid();

#if 0
  (void) sprintf(pidfilename, "%s/%s.pid", pidpath, ifname);

  /* write pid to file */

  if ((pidfile = fopen(pidfilename, "w")) != NULL) {
    fprintf(pidfile, "%d\n", pid);
    (void) fclose(pidfile);
  }
#endif

  sifframing(ifunit, slip);

  if (dialup)
    sifbringup(DP_IN_PROGRESS);

#else /* SOLARIS_2 */

  /*
   * Push the Async/FCS module and VJC module onto this stream.
   */
  if(ioctl(fd, I_PUSH, syncmode ? "pppsync" : "pppasync") < 0) {
    syslog(LOG_ERR, "ioctl(I_PUSH, \"ppp%ssync\"): %m", syncmode ? "" : "a");
    exit(1);
  }
  str_push_count++;
  if (ioctl(fd, I_PUSH, "vjc") < 0) {
    syslog(LOG_ERR, "ioctl(I_PUSH, \"vjc\"): %m");
    exit(1);
  }
  str_push_count++;

  /*
   * Now, open the dp device and set it up to accept LCP and NCP packets
   */
  if ((dpfd = open(DEVDP, O_RDWR)) < 0) {
    syslog(LOG_ERR, "Can't open \"%s\", %m", DEVDP);
    exit(1);
  }

  /*
   * Attach to the specified PPA (unit), and bind to the packets
   * of interest, namely packets directed towards the LCP and NCP
   * protocol families.
   */
  if ((dlattachreq(dpfd, ifunit) == -1) ||
      dlokack(dpfd, buf) < 0) {
    syslog(LOG_ERR, "Can't attach to dp PPA %d, %m\n", ifunit);
    exit(1);
  }

  if (!dialup) {
    dp_timeo_t dpt;

     dpt.dt_active =
     dpt.dt_active_c =
     dpt.dt_active_u =
     dpt.dt_waiting =
     dpt.dt_failcall = DPT_NO_TIMEOUT;

    if ((strioctl(dpfd, DPIOCSTIMEOS, -1, sizeof(dpt), &dpt)) < 0) {
      syslog(LOG_ERR, "ioctl(DPIOCSTIMEOS): %m");
      exit(1);
    }
  }
  sap = DP_PROTO_SAP(PPP_LCP);
  if ((dlbindreq(dpfd, sap, 0, DL_CLDLS, 0, 0) == -1) ||
      dlbindack(dpfd, buf) < 0) {
    syslog(LOG_ERR, "Can't bind to dp DLSAP %x, %m\n", DP_PROTO_SAP(PPP_LCP));
    exit(1);
  }

  sap = DP_PROTO_SAP(PPP_NCP);
  if ((dlsubsbindreq(dpfd, &sap, sizeof(sap), DL_PEER_BIND) == -1) ||
      dlsubsbindack(dpfd, buf)) {
    syslog(LOG_ERR, "Can't bind to dp DLSAP %x, %m\n", DP_PROTO_SAP(PPP_NCP));
    exit(1);
  }

  /*
   * Finally, link this module to the dp device.
   */
  if((dpmuxid = ioctl(dpfd, I_LINK, fd)) < 0) {
    syslog(LOG_ERR, "ioctl(I_LINK): %m");
    exit(1);
  }

  if(fcntl(dpfd, F_SETFL, O_NDELAY) < 0) {
    syslog(LOG_ERR, "fcntl(F_SETFL, O_NDELAY): %m");
    exit(1);
  }
  /* read mode, message non-discard mode */
  if(ioctl(dpfd, I_SRDOPT, RMSGN) < 0) {
    syslog(LOG_ERR, "ioctl(I_SRDOPT, RMSGN): %m");
    exit(1);
  }
  /* Flush any waiting messages, or we'll never get SIGPOLL */
  if(ioctl(dpfd, I_FLUSH, FLUSHRW) < 0) {
    syslog(LOG_ERR, "ioctl(I_FLUSH, FLUSHRW): %m");
    exit(1);
  }

#if	0
  /*
   * Find out which interface we were given.
   */
  /* in streams, dp_if handles this ioctl */
  if (ioctl(fd, DPIOCGIFUNIT, &ifunit) < 0) {
    syslog(LOG_ERR, "ioctl(DPIOCGIFUNIT): %m");
    exit(1);
  }
#endif
  sprintf(ifname, "dp%d", ifunit);
  syslog(LOG_NOTICE, "Using unit %s", ifname);
  pid = getpid();

#if 0
  (void) sprintf(pidfilename, "%s/%s.pid", pidpath, ifname);

  /* write pid to file */

  if ((pidfile = fopen(pidfilename, "w")) != NULL) {
    fprintf(pidfile, "%d\n", pid);
    (void) fclose(pidfile);
  }
#endif

  sifframing(ifunit, slip);

  if (dialup) {
    sifbringup(DP_IN_PROGRESS);
  }

#endif /* SOLARIS_1 or SOLARIS_2 */

  hostname_len = (u_char) strlen(hostname);

  MAINDEBUG((LOG_DEBUG, "hostname = %s", hostname))
      
#if 0
#ifdef SETSID
  if (default_device) {
    int id = tcgetpgrp(fd);
    if (id != pgrpid) {
      MAINDEBUG((LOG_WARNING, "warning: pppd is not the leader of a foreground process group"))
    }
  }
  else
    if (tcsetpgrp(fd, pgrpid) < 0) {
      syslog(LOG_ERR, "tcsetpgrp(): %m");
      exit(1);
    }
#else
  if (ioctl(fd, TIOCSPGRP, &pid) < 0) {
    syslog(LOG_ERR, "ioctl(TIOCSPGRP): %m");
    exit(1);
  }
#endif
#endif
  
  /*
   * Compute mask of all interesting signals and install signal handlers
   * for each.  Only one signal handler may be active at a time.  Therefore,
   * all other signals should be masked when any handler is executing.
   */
  sigemptyset(&mask);
  sigaddset(&mask, SIGHUP);
  sigaddset(&mask, SIGINT);
  sigaddset(&mask, SIGALRM);
  sigaddset(&mask, SIGPOLL);	/* = SIGIO in SOLARIS_1 */

  /*
   * Block all signals until we get things set up here.
   */
  sigprocmask(SIG_BLOCK, &mask, NULL);	/* Block signals now */

#if	0
  /*
   * How do we get SIGTERMs?
   */
  if (ioctl(fd, TIOCSPGRP, &pid) < 0) {
    syslog(LOG_ERR, "ioctl(TIOCSPGRP): %m");
    exit(1);
  }
#endif

  sa.sa_handler = hup;		/* Hangup */
  sa.sa_mask = mask;
  sa.sa_flags = 0;
  if (sigaction(SIGHUP, &sa, NULL)) {
    syslog(LOG_ERR, "sigaction(SIGHUP)");
    exit(1);
  }
  sa.sa_handler = intr;		/* Interrupt */
  sa.sa_mask = mask;
  sa.sa_flags = 0;
  if (sigaction(SIGINT, &sa, NULL)) {
    syslog(LOG_ERR, "sigaction(SIGINT)");
    exit(1);
  }
  sa.sa_handler = term;		/* Terminate */
  sa.sa_mask = mask;
  sa.sa_flags = 0;
  if (sigaction(SIGTERM, &sa, NULL)) {
    syslog(LOG_ERR, "sigaction(SIGTERM)");
    exit(1);
  }
  sa.sa_handler = alrm;		/* Timeout */
  sa.sa_mask = mask;
  sa.sa_flags = 0;
  if (sigaction(SIGALRM, &sa, NULL)) {
    syslog(LOG_ERR, "sigaction(SIGALRM)");
    exit(1);
  }
  sa.sa_handler = io;			/* Input available */
  sa.sa_mask = mask;
  sa.sa_flags = 0;
  if (sigaction(SIGPOLL, &sa, NULL)) {
    syslog(LOG_ERR, "sigaction(SIGPOLL)");
    exit(1);
  }
  (void) signal(SIGUSR1, incdebug);	/* Increment debug flag */
  (void) signal(SIGUSR2, nodebug);	/* Reset debug flag */

#ifdef SOLARIS_1
  /*
   * Record initial device flags, then set device to cause SIGIO
   * signals to be generated.
   */
  if ((initfdflags = fcntl(fd, F_GETFL)) == -1) {
    syslog(LOG_ERR, "fcntl(F_GETFL): %m");
    exit(1);
  }
  if (fcntl(fd, F_SETFL, FNDELAY | FASYNC) == -1) {
    syslog(LOG_ERR, "fcntl(F_SETFL, FNDELAY | FASYNC): %m");
    exit(1);
  }
#else /* SOLARIS_2 */
  /*
   * Enable reception of signals generated by the streams driver.
   */
  if (ioctl(dpfd, I_SETSIG, S_INPUT|S_RDNORM|S_RDBAND|S_HIPRI|S_HANGUP) < 0) {
    syslog(LOG_ERR, "ioctl(I_SETSIG, S_RDNORM|S_RDBAND|S_HIPRI|S_HANGUP): %m");
    exit(1);
  }
#endif

  MAINDEBUG((LOG_INFO, "connect: %s %s", ifname, devname))
  
  
  /*
   * Start opening the connection, and  wait for
   * incoming signals (reply, timeout, etc.).
   */

  
  if(slip) {

    sifdown(ifunit);
    sifsapdisable(DP_PROTO_SAP(PPP_NET));
    ppp_send_config(ifunit, lcp_wantoptions[0].mru, 0xffffffff, 0, 0, 0);
 
    if (ipcp_wantoptions[0].neg_addr && ipcp_wantoptions[0].ouraddr &&
	  ipcp_wantoptions[0].hisaddr)
      sifaddr(ifunit, ipcp_wantoptions[0].ouraddr, ipcp_wantoptions[0].hisaddr, 
	GetMask(ipcp_wantoptions[0].ouraddr));
    sifvjcomp(ifunit, ipcp_wantoptions[0].neg_vj,
		ipcp_wantoptions[0].cflag,
		ipcp_wantoptions[0].maxslotindex);
    sifbringup(DP_SUCCESS);
    sifup(ifunit);
    sifsapenable(PPP_IP);
  } else {
    lcp_lowerup(0);			/* XXX Well, sort of... */
    lcp_open(0);				/* Start protocol */
  }

  for (phase = PHASE_ESTABLISH; phase != PHASE_DEAD; ) {
    sigset_t zmask;
    sigemptyset(&zmask);

    sigsuspend(&zmask);			/* Wait for next signal */
  }

  quit();
}


/*
 * quit - Clean up state and exit.
 */
void 
quit()
{
  static int entered = 0;

  if (entered) {
	MAINDEBUG ((LOG_DEBUG, "quit(): reentered"))
	return;
  }

  entered = 1;

#ifdef undef
  /*
   * this is probably not needed and seems to break things.  ifdef'ed
   * out for now - gmc
   */

  /*
   * the I_FLUSH ioctl should return ENXIO if carrier is missing on the 
   * line, but due to a typo, it returns EINVAL instead.  We ignore
   * both.
   */
  if((ioctl(fd, I_FLUSH, FLUSHRW) < 0) &&
     (errno != EINVAL) && (errno != ENXIO)) { 
    syslog(LOG_ERR, "quit ioctl(I_FLUSH,FLUSHRW)");
  }
#endif
#if	0
  str_restore();
#endif
  if (slip) {
    sifdown(ifunit);
    if (ipcp_wantoptions[0].neg_addr && ipcp_wantoptions[0].ouraddr &&
	  ipcp_wantoptions[0].hisaddr)
      cifaddr(ifunit, ipcp_wantoptions[0].ouraddr, ipcp_wantoptions[0].hisaddr);
  }
  /* pmg - this shouldn't close here, since on_exit calls cleanup,
   *       which really confuses matters...
   */
  /* close(fd); */
  str_restore();
#if 0
  if(pidfilename[0]) {
    if (unlink(pidfilename) < 0) 
      syslog(LOG_WARNING, "unable to unlink pid file: %m");
    pidfilename[0]='\0';
  }
#endif
  exit(0);
}


static struct callout *callout = NULL;		/* Callout list */
static struct timeval schedtime;		/* Time last timeout was set */


/*
 * timeout - Schedule a timeout.
 *
 * Note that this timeout takes the number of seconds, NOT hz (as in
 * the kernel).
 */
void
timeout(func, arg, time)
void (*func)();
caddr_t arg;
int time;
{
  struct itimerval itv;
  struct callout *newp, **oldpp;
  
  MAINDEBUG((LOG_DEBUG, "Timeout %x:%x in %d seconds.",
	    (int) func, (int) arg, time))
  
  /*
   * Allocate timeout.
   */
  if ((newp = (struct callout *) malloc(sizeof(struct callout))) == NULL) {
    syslog(LOG_ERR, "Out of memory in timeout()!");
    exit(1);
  }
  newp->c_arg = arg;
  newp->c_func = func;
  
  /*
   * Find correct place to link it in and decrement its time by the
   * amount of time used by preceding timeouts.
   */
  for (oldpp = &callout;
       *oldpp && (*oldpp)->c_time <= time;
       oldpp = &(*oldpp)->c_next)
    time -= (*oldpp)->c_time;
  newp->c_time = time;
  newp->c_next = *oldpp;
  if (*oldpp)
    (*oldpp)->c_time -= time;
  *oldpp = newp;
  
  /*
   * If this is now the first callout then we have to set a new
   * itimer.
   */
  if (callout == newp) {
    itv.it_interval.tv_sec = itv.it_interval.tv_usec =
      itv.it_value.tv_usec = 0;
    itv.it_value.tv_sec = callout->c_time;
    MAINDEBUG((LOG_DEBUG, "Setting itimer for %d seconds.",
	      itv.it_value.tv_sec))
    if (setitimer(ITIMER_REAL, &itv, NULL)) {
      syslog(LOG_ERR, "setitimer(ITIMER_REAL)");
      exit(1);
    }
    if (gettimeofday(&schedtime, NULL)) {
      syslog(LOG_ERR, "gettimeofday");
      exit(1);
    }
  }
}


/*
 * untimeout - Unschedule a timeout.
 */
void
untimeout(func, arg)
void (*func)();
caddr_t arg;
{
  
  struct itimerval itv;
  struct callout **copp, *freep;
  int reschedule = 0;
  
  MAINDEBUG((LOG_DEBUG, "Untimeout %x:%x.", (int) func, (int) arg))
  
  /*
   * If the first callout is unscheduled then we have to set a new
   * itimer.
   */
  if (callout &&
      callout->c_func == func &&
      callout->c_arg == arg)
    reschedule = 1;
  
  /*
   * Find first matching timeout.  Add its time to the next timeouts
   * time.
   */
  for (copp = &callout; *copp; copp = &(*copp)->c_next)
    if ((*copp)->c_func == func &&
	(*copp)->c_arg == arg) {
      freep = *copp;
      *copp = freep->c_next;
      if (*copp)
	(*copp)->c_time += freep->c_time;
      (void) free((char *) freep);
      break;
    }
  
  if (reschedule) {
    itv.it_interval.tv_sec = itv.it_interval.tv_usec =
      itv.it_value.tv_usec = 0;
    itv.it_value.tv_sec = callout ? callout->c_time : 0;
    MAINDEBUG((LOG_DEBUG, "Setting itimer for %d seconds.",
	      itv.it_value.tv_sec))
    if (setitimer(ITIMER_REAL, &itv, NULL)) {
      syslog(LOG_ERR, "setitimer(ITIMER_REAL)");
      exit(1);
    }
    if (gettimeofday(&schedtime, NULL)) {
      syslog(LOG_ERR, "gettimeofday");
      exit(1);
    }
  }
}


/*
 * adjtimeout - Decrement the first timeout by the amount of time since
 * it was scheduled.
 */
void
adjtimeout()
{
  struct timeval tv;
  int timediff;
  
  if (callout == NULL)
    return;
  /*
   * Make sure that the clock hasn't been warped dramatically.
   * Account for recently expired, but blocked timer by adding
   * small fudge factor.
   */
  if (gettimeofday(&tv, NULL)) {
    syslog(LOG_ERR, "gettimeofday");
    exit(1);
  }
  timediff = tv.tv_sec - schedtime.tv_sec;
  if (timediff < 0 ||
      timediff > callout->c_time + 1)
    return;
  
  callout->c_time -= timediff;	/* OK, Adjust time */
}


/*
 * output - Output PPP packet.
 */
void
output(unit, p, len)
int unit;
u_char *p;
int len;
{
#ifdef SOLARIS_1

  struct strbuf	str;
  if (unit != 0) {
    MAINDEBUG((LOG_WARNING, "output: unit != 0!"))
    abort();
  }

  if (debug)
    log_packet(p, len, "sent ");

  if (debug > 2) {char dbuf[10*1024];
   int i;
   strcpy(dbuf, "");
   for (i=0; i<len; i++)
	sprintf(dbuf+strlen(dbuf), "%02x", p[i]);
   MAINDEBUG((LOG_DEBUG, "output: %s", dbuf))
  }
  str.len = len;
  str.buf = (caddr_t) p;
  if(putmsg(fd, NULL, &str, 0) < 0) {
    MAINDEBUG((LOG_ERR, "putmsg"));
    exit(1);
  }
  free(p);

#else /* SOLARIS_2 */

  phdr_t *ph;
  u_long dlsap;
  if (unit != 0) {
    MAINDEBUG((LOG_WARNING, "output: unit != 0!"))
    abort();
  }

  if (debug)
    log_packet(p, len, "sent ");

  ph = (phdr_t *)p;
  p += sizeof(*ph);
  len -= sizeof(*ph);
  dlsap = DP_DLSAP(ph->ph_address, ph->ph_control, ntohs(ph->ph_protocol));

  if (dlunitdatareq(dpfd, &dlsap, sizeof(dlsap), 0, 0, p, len) == -1) {
    MAINDEBUG((LOG_DEBUG, "dlunitdatareq, %m"));
    exit(1);
  }

#endif /* SOLARIS_1 or SOLARIS_2 */
}


/*
 * hup - Catch SIGHUP signal.
 *
 * Indicates that the physical layer has been disconnected.
 */
/*ARGSUSED*/
static void
hup(sig)
int sig;
{
  MAINDEBUG((LOG_NOTICE, "Hangup"))
  adjtimeout();		/* Adjust timeouts */
  if (!slip) {
    if (dialup)
      lcp_close(0);
    else
      lcp_lowerdown(0);		/* Reset connection */
  }
  if(!lcp_wantoptions[0].restart)
	quit();
}


/*
 * term - Catch SIGTERM signal.
 *
 * Indicates that we should initiate a graceful disconnect and exit.
 */
/*ARGSUSED*/
static void
term(sig)
int sig;
{
  syslog(LOG_NOTICE, "Terminate signal received.");
  adjtimeout();		/* Adjust timeouts */
  if (slip)
    quit();
  else
    lcp_close(0);		/* Close connection */
}


/*
 * intr - Catch SIGINT signal (DEL/^C).
 *
 * Indicates that we should initiate a graceful disconnect and exit.
 */
/*ARGSUSED*/
static void
intr(sig)
int sig;
{
  syslog(LOG_NOTICE, "Interrupt received.  Exiting.");
  adjtimeout();		/* Adjust timeouts */
  if (slip)
    quit();
  else
    lcp_close(0);		/* Close connection */
}


/*
 * alrm - Catch SIGALRM signal.
 *
 * Indicates a timeout.
 */
/*ARGSUSED*/
static void
alrm(sig)
int sig;
{
  struct itimerval itv;
  struct callout *freep;
  
  MAINDEBUG((LOG_DEBUG, "Alarm"))
  
  /*
   * Call and free first scheduled timeout and any that were scheduled
   * for the same time.
   */
  while (callout) {
    freep = callout;	/* Remove entry before calling */
    callout = freep->c_next;
    (*freep->c_func)(freep->c_arg);
    (void) free((char *) freep);
    if (callout && callout->c_time)
      break;
  }
  
  /*
   * Set a new itimer if there are more timeouts scheduled.
   */
  if (callout) {
    itv.it_interval.tv_sec = itv.it_interval.tv_usec =
      itv.it_value.tv_usec = 0;
    itv.it_value.tv_sec = callout->c_time;
    MAINDEBUG((LOG_DEBUG, "Setting itimer for %d seconds.",
	      itv.it_value.tv_sec))
    if (setitimer(ITIMER_REAL, &itv, NULL)) {
      syslog(LOG_ERR, "setitimer(ITIMER_REAL)");
      exit(1);
    }
    if (gettimeofday(&schedtime, NULL)) {
      syslog(LOG_ERR, "gettimeofday");
      exit(1);
    }
  }
}

#if	0

#define	MAX_SAVE	128

char *malloc_save[MAX_SAVE];
char *free_save[MAX_SAVE];

int nma, nfr;

static
char *
Malloc(s)
uint s;
{
    char *p;

    p = malloc(s);
    malloc_save[nma++] = p;
    return p;
}

static void
Free(p)
char *p;
{
    free_save[nfr++] = p;
    free(p);
}
#endif

/*
 * io - Catch SIGPOLL signal.
 *
 * Indicates that incoming data is available.
 */
/*ARGSUSED*/
static void
io(sig)
int sig;
{
#ifdef SOLARIS_1

  int len, i;
  u_char *p, *pstart;
  u_short protocol;
  fd_set fdset;
  struct timeval notime;
  int ready;

  struct strbuf str;
  MAINDEBUG((LOG_DEBUG, "IO signal received"))
  adjtimeout();		/* Adjust timeouts */

  /* we do this to see if the SIGIO handler is being invoked for input */
  /* ready, or for the socket buffer hitting the low-water mark. */

  notime.tv_sec = 0;
  notime.tv_usec = 0;

  FD_ZERO(&fdset);
  FD_SET(fd, &fdset);
  
  if ((ready = select(32, &fdset, (fd_set *) NULL, (fd_set *) NULL,
		      &notime)) == -1) {
    syslog(LOG_ERR, "Error in io() select: %m");
    exit(1);
  }
    
  if (ready == 0) {
    MAINDEBUG((LOG_DEBUG, "IO non-input ready SIGIO occured."));
    return;
  }

  /* Yup, this is for real */
  for (;;) {			/* Read all available packets */
    pstart = p = (u_char *) malloc(MTU + DLLHEADERLEN);
    str.maxlen = MTU+DLLHEADERLEN;
    str.buf = (caddr_t) p;
    i = 0;
    len = getmsg(fd, NULL, &str, &i);
    if(len < 0) {
      if(errno == EAGAIN || errno == EWOULDBLOCK) {
        free(pstart);
	return;
      }
      syslog(LOG_ERR, "getmsg(fd) %m");
      exit(1);
    }
    else if(len) 
      MAINDEBUG((LOG_DEBUG, "getmsg returns with length 0x%x",len))
    
    if(str.len < 0) {
      MAINDEBUG((LOG_DEBUG, "getmsg short return length %d",
		str.len))
      free(pstart);
      return;
    }
    
    len = str.len;
      if (len == 0) {
	syslog(LOG_ERR, "End of file on fd!");
	exit(1);
      }
    
    if (len < DLLHEADERLEN) {
      MAINDEBUG((LOG_INFO, "io(): Received short packet."))
      free(pstart);
      return;
    }
    
    if (debug)
      log_packet(p, len, "rcvd ");

    p += 2;				/* Skip address and control */
    GETSHORT(protocol, p);
    len -= DLLHEADERLEN;
    
    /*
     * Toss all non-LCP packets unless LCP is OPEN.
     */
    if (protocol != LCP && lcp_fsm[0].state != OPENED) {
      MAINDEBUG((LOG_INFO, "io(): Received non-LCP packet and LCP is not in open state."))
      free(pstart);
      return;
    }
    
    /*
     * Upcall the proper protocol input routine.
     */
    for (i = 0; i < sizeof (prottbl) / sizeof (struct protent); i++)
      if (prottbl[i].protocol == protocol) {
	(*prottbl[i].input)(0, p, len);
	break;
      }
    
    if (i == sizeof (prottbl) / sizeof (struct protent)) {
      syslog(LOG_WARNING, "input: Unknown protocol (%x) received!",
	     protocol);
      p -= DLLHEADERLEN;
      len += DLLHEADERLEN;
      lcp_sprotrej(0, p, len);
    }
  }

#else /* SOLARIS_2 */

    int len, i;
    u_char *p, *pstart;
    u_short protocol;
    fd_set fdset;
    struct timeval notime;
    int ready;
    ulong dlsap;
    union DL_primitives *dl;
    dl_unitdata_ind_t *dui;
    dl_signal_ind_t *dsi;
    dl_uderror_ind_t *due;
    char ctlbuf[128];
    phdr_t *ph;
    struct pollfd pfd;
    int nmsgs = 0;

    struct strbuf ctl, data;

    MAINDEBUG((LOG_DEBUG, "POLL signal received"))
    adjtimeout();		/* Adjust timeouts */

    /* Yup, this is for real */
    for (;;) {			/* Read all available packets */
	/* we do this to see if the SIGPOLL handler is being invoked for 
	   input ready, or for the socket buffer hitting the low-water mark. */

	pfd.fd = dpfd;
	pfd.events = POLLRDNORM|POLLRDBAND|POLLPRI;
	pfd.revents = 0;

	if (poll(&pfd, 1, 0) < 0) {
	    syslog(LOG_ERR, "Error in io() poll: %m");
	    exit(1);
	}

	if (!pfd.revents) {
	    if (nmsgs == 0)
		MAINDEBUG((LOG_DEBUG, "IO non-input ready SIGPOLL occured."));
	    return;
	}

	if (pfd.revents & (POLLERR|POLLNVAL)) {
	    syslog(LOG_DEBUG, "io: poll returns %s%s",
	    (pfd.revents & POLLERR)  ? "POLLERR "  : "",
	    (pfd.revents & POLLNVAL) ? "POLLNVAL " : "");
	}

	if (pfd.revents & POLLHUP) {
	    MAINDEBUG((LOG_DEBUG, "io(): POLLHUP"));
	    hup(SIGHUP);
	    return;
	}

	nmsgs++;
	pstart = (u_char *) malloc(DLLHEADERLEN + MTU);
	p = pstart + DLLHEADERLEN;

	ctl.maxlen = sizeof(ctlbuf);
	ctl.buf = ctlbuf;

	data.maxlen = MTU;
	data.buf = (caddr_t)p;

	i = 0;

	if (strgetmsg(dpfd, &ctl, &data, &i, "io") < 0) {
	    if (errno == EAGAIN || errno == EWOULDBLOCK) {
		free(pstart);
		return;
	    }
	    syslog(LOG_ERR, "strgetmsg(dpfd) %m");
	    exit(1);
	}
	MAINDEBUG((LOG_DEBUG, "io(): Received message"));

	if (ctl.len < sizeof(u_long)) {
	    MAINDEBUG((LOG_DEBUG, "io(): getmsg short return length %d",
		       ctl.len))
	    free(pstart);
	    return;
	}

	dl = (union DL_primitives *)ctl.buf;
	switch (dl->dl_primitive) {
	 case DL_UNITDATA_IND:
	    dui = &dl->unitdata_ind;
	    break;

	 case DL_UDERROR_IND:
	    due = &dl->uderror_ind;
	    errno = due->dl_unix_errno;
	    syslog(LOG_ERR, "io(): DL_UDERROR_IND: error 0x%x, %m",
		   due->dl_errno);
	    free(pstart);
	    continue;

	 case DL_SIGNAL_IND:
	    dsi = (dl_signal_ind_t *)&dl->dl_primitive;
	    if (ctl.len < sizeof(*dsi))
		syslog(LOG_ERR, "io(): DL_SIGNAL_IND: short message");
	    else
		kill(getpid(), dsi->dl_signal);
	    free(pstart);
	    continue;

	 default:
	    syslog(LOG_ERR, "io(): unknown DLPI message: 0x%x",
		   dl->dl_primitive);
	    free(pstart);
	    continue;
	}

	if (ctl.len < sizeof(*dui)) {
	    MAINDEBUG((LOG_DEBUG, "io(): getmsg short return length %d",
		      ctl.len))
	    free(pstart);
	    return;
	}

	if (dui->dl_dest_addr_length != sizeof(u_long)) {
	    MAINDEBUG((LOG_DEBUG,
		      "io(): DL_UNITDATA_IND bad dest address length %d",
		      dui->dl_dest_addr_length))
	    free(pstart);
	    return;
	}

	dlsap = *(ulong *)(((char *)dui) + dui->dl_dest_addr_offset);
	protocol = DP_DLSAP_PROTO(dlsap);

	if (data.len < 0) {
	    MAINDEBUG((LOG_DEBUG, "getmsg short return data length %d",
		       data.len))
	    free(pstart);
	    return;
	}

	len = data.len;

	if (debug > 2) {
	    char dbuf[10*1024];
	    int i;
	    strcpy(dbuf, "");
	    for (i=0; i<len; i++)
		sprintf(dbuf+strlen(dbuf), "%02x", p[i]);
	    MAINDEBUG((LOG_DEBUG, "input: %s", dbuf))
	}

	if (debug) {
	    ph = (phdr_t *) (p - DLLHEADERLEN);
	    ph->ph_address  = DP_DLSAP_ADDR(dlsap);
	    ph->ph_control  = DP_DLSAP_CTL(dlsap);
	    ph->ph_protocol = htons(DP_DLSAP_PROTO(dlsap));
	    log_packet(p - DLLHEADERLEN, len + DLLHEADERLEN, "rcvd ");
	}

	/*
	* Toss all non-LCP packets unless LCP is OPEN.
	*/
	if (protocol != LCP && lcp_fsm[0].state != OPENED) {
	    MAINDEBUG((LOG_INFO,
	"io(): Received non-LCP packet (%04x) and LCP is not in open state.",
		      protocol))
	    free(pstart);
	    return;
	}

	/*
	 * Upcall the proper protocol input routine.
	 */
	for (i = 0; i < sizeof (prottbl) / sizeof (struct protent); i++)
	    if (prottbl[i].protocol == protocol) {
		(*prottbl[i].input)(0, p, len);
		break;
	    }

	if (i == sizeof (prottbl) / sizeof (struct protent)) {
	    syslog(LOG_WARNING, "input: Unknown protocol (%x) received!",
		   protocol);
	    p -= DLLHEADERLEN;
	    len += DLLHEADERLEN;
	    ph = (phdr_t *)p;
	    ph->ph_address  = DP_DLSAP_ADDR(dlsap);
	    ph->ph_control  = DP_DLSAP_CTL(dlsap);
	    ph->ph_protocol = htons(DP_DLSAP_PROTO(dlsap));
	    lcp_sprotrej(0, p, len);
	}
	free(pstart);
    }

#endif /* SOLARIS_1 or SOLARIS_2 */
}

/*
 * cleanup - clean_up before we exit
 */
/* ARGSUSED */
static void
#ifdef SOLARIS_1
cleanup(status, arg)
int status;
caddr_t arg;
#else
cleanup()
#endif
{
  adjtimeout();
  if (!slip)
    lcp_lowerdown(0);
  quit();
#if 0
  if (unlink(pidfilename) < 0) 
    syslog(LOG_WARNING, "unable to unlink pid file: %m");
#endif
}

/*
 * demuxprotrej - Demultiplex a Protocol-Reject.
 */
void
demuxprotrej(unit, protocol)
int unit;
int protocol;
{
  int i;
  
  /*
   * Upcall the proper Protocol-Reject routine.
   */
  for (i = 0; i < sizeof (prottbl) / sizeof (struct protent); i++)
    if (prottbl[i].protocol == protocol) {
      (*prottbl[i].protrej)(unit);
      return;
    }
  syslog(LOG_WARNING, "demuxprotrej: Unrecognized Protocol-Reject for protocol %d!", protocol);
}


/*
 * incdebug - Catch SIGUSR1 signal.
 *
 * Increment debug flag.
 */
/*ARGSUSED*/
static void
incdebug(sig)
int sig;
{
  MAINDEBUG((LOG_NOTICE, "Debug turned ON, Level %d",debug))
  setlogmask(LOG_UPTO(LOG_DEBUG));
  debug++;
}


/*
 * nodebug - Catch SIGUSR2 signal.
 *
 * Turn off debugging.
 */
/*ARGSUSED*/
static void
nodebug(sig)
int sig;
{
  setlogmask(LOG_UPTO(LOG_WARNING));
  debug = 0;
}


/*
 * setdebug - Set debug (command line argument).
 */
static int
setdebug(argcp, argvp)
int *argcp;
char ***argvp;
{
  debug++;
  setlogmask(LOG_UPTO(LOG_DEBUG));
  --*argcp, ++*argvp;
  return (1);
}

/*
 * noopt - Disable all options.
 */
static int
noopt(argcp, argvp)
int *argcp;
char ***argvp;
{
  memset((char *) &lcp_wantoptions[0],   0, sizeof (struct lcp_options));
  memset((char *) &lcp_allowoptions[0],  0, sizeof (struct lcp_options));
  memset((char *) &ipcp_wantoptions[0],  0, sizeof (struct ipcp_options));
  memset((char *) &ipcp_allowoptions[0], 0, sizeof (struct ipcp_options));
  --*argcp, ++*argvp;
  return (1);
}


/*
 * setdialup - Set dialup
 */
static int
setdialup(argcp, argvp)
int *argcp;
char ***argvp;
{
  dialup++;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * setdynamic - Allow dynamic addresses
 */
static int
setdynamic(argcp, argvp)
int *argcp;
char ***argvp;
{
  dynamic++;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * noaccomp - Disable Address/Control field compression negotiation.
 */
static int
noaccomp(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].neg_accompression = 0;
  lcp_allowoptions[0].neg_accompression = 0;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * noasyncmap - Disable async map negotiation.
 */
static int
noasyncmap(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].neg_asyncmap = 0;
  lcp_allowoptions[0].neg_asyncmap = 0;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * noipaddr - Disable IP address negotiation.
 */
static int
noipaddr(argcp, argvp)
int *argcp;
char ***argvp;
{
  ipcp_wantoptions[0].neg_addr = 0;
  ipcp_allowoptions[0].neg_addr = 0;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * nomagicnumber - Disable magic number negotiation.
 */
static int
nomagicnumber(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].neg_magicnumber = 0;
  lcp_allowoptions[0].neg_magicnumber = 0;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * nomru - Disable mru negotiation.
 */
static int
nomru(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].neg_mru = 0;
  lcp_allowoptions[0].neg_mru = 0;
  --*argcp, ++*argvp;
  return (1);
}

/*
 * setmru - Set MRU for negotiation.
 */
static int
setmru(argcp, argvp)
int *argcp;
char ***argvp;
{
  --*argcp, ++*argvp;
  lcp_wantoptions[0].mru = atoi(**argvp);
  --*argcp, ++*argvp;
  return (1);
}


/*
 * nopcomp - Disable Protocol field compression negotiation.
 */
static int
nopcomp(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].neg_pcompression = 0;
  lcp_allowoptions[0].neg_pcompression = 0;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * setpassive - Set passive mode.
 */
static int
setpassive(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].passive = 1;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * setsilent - Set silent mode.
 */
static int
setsilent(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].silent = 1;
  --*argcp, ++*argvp;
  return (1);
}

/*
 * noupap - Disable UPAP authentication.
 */
static int
noupap(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_allowoptions[0].neg_upap = 0;
  --*argcp, ++*argvp;
  return (1);
}

/*
 * requpap - Require UPAP authentication.
 */
static int
requpap(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].neg_upap = 1;
  lcp_allowoptions[0].neg_upap = 0;
  auth_required = 1;
  --*argcp, ++*argvp;
  return (1);
}

/*
 * setunit - Set Unit
 */
static int
setunit(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    ifunit = atoi(**argvp);
    --*argcp, ++*argvp;
    return (1);
}

/*
 * nochap - Disable CHAP authentication.
 */
static int
nochap(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_allowoptions[0].neg_chap = 0;
  --*argcp, ++*argvp;
  return (1);
}


/*
 * reqchap - Require CHAP authentication.
 */
static int
reqchap(argcp, argvp)
int *argcp;
char ***argvp;
{
  lcp_wantoptions[0].neg_chap = 1;
  lcp_allowoptions[0].neg_chap = 0;
  auth_required = 1;
  --*argcp, ++*argvp;
  return (1);
}

/*
 *	setvjmode - Set vj compression mode
 */

static int
setvjmode(argcp, argvp)
int *argcp;
char ***argvp;
{
  extern int ipcp_vj_mode;
  
  --*argcp, ++*argvp;
  
  if (!strcmp(**argvp, "rfc1172") ||
      !strcmp(**argvp, "old")) {		/* "old" mode */
    ipcp_vj_setmode(IPCP_VJMODE_OLD);
  }
  
  else if (!strcmp(**argvp, "pre1332") ||
	   !strcmp(**argvp, "current")) {	/* "current" mode*/
    ipcp_vj_setmode(IPCP_VJMODE_RFC1172);
  }
  
  else if (!strcmp(**argvp, "rfc1332") ||
	   !strcmp(**argvp, "draft")) {		/* "draft" mode (default)*/
    ipcp_vj_setmode(IPCP_VJMODE_RFC1332);
  }
  else {
    MAINDEBUG((LOG_WARNING, "Unknown vj compression mode %s. Defaulting to rfc1332.", **argvp))
    ipcp_vj_setmode(IPCP_VJMODE_RFC1332);
  }
  --*argcp, ++*argvp;
  
  return (1);
}
/*
 *	setnovj - disable vj compression
 */

static int
setnovj(argcp, argvp)
int *argcp;
char ***argvp;
{
  extern int ipcp_vj_mode;
  
  --*argcp, ++*argvp;
  ipcp_wantoptions[0].neg_vj =
  ipcp_allowoptions[0].neg_vj = 0;
  
  return (1);
}
/*
 *	setnovjccomp - disable VJ connection-ID compression
 */

static int
setnovjccomp(argcp, argvp)
int *argcp;
char ***argvp;
{
  --*argcp, ++*argvp;
  ipcp_wantoptions[0].cflag =
  ipcp_allowoptions[0].cflag = 0;
  
  return (1);
}
/*
 *	 setvjslots - set maximum number of connection slots for VJ compression
 */

static int
setvjslots(argcp, argvp)
int *argcp;
char ***argvp;
{
  int slots;
  --*argcp, ++*argvp;
  slots = atoi(**argvp);
  --*argcp, ++*argvp;
  if (slots >= VJC_MIN_STATES && slots <= VJC_MAX_STATES) {
      ipcp_wantoptions[0].maxslotindex =
      ipcp_allowoptions[0].maxslotindex = slots - 1;
  }
  else { 
	fprintf(stderr, "%s: vjslots must be between %d and %d\n", progname,
		VJC_MIN_STATES, VJC_MAX_STATES);
	return 0;
  }
  
  return (1);
}

/*
 * setdomain - Set domain name to append to hostname 
 */
static int
setdomain(argcp, argvp)
int *argcp;
char ***argvp;
{
  
  --*argcp, ++*argvp;

  strcat(hostname, **argvp);
  hostname_len = strlen(hostname);

  --*argcp, ++*argvp;

  return (1);
}

/*
 * Valid speeds.
 */
struct speed {
  int speed_int, speed_val;
} speeds[] = {
#ifdef B50
  { 50, B50 },
#endif
#ifdef B75
  { 75, B75 },
#endif
#ifdef B110
  { 110, B110 },
#endif
#ifdef B150
  { 150, B150 },
#endif
#ifdef B200
  { 200, B200 },
#endif
#ifdef B300
  { 300, B300 },
#endif
#ifdef B600
  { 600, B600 },
#endif
#ifdef B1200
  { 1200, B1200 },
#endif
#ifdef B1800
  { 1800, B1800 },
#endif
#ifdef B2000
  { 2000, B2000 },
#endif
#ifdef B2400
  { 2400, B2400 },
#endif
#ifdef B3600
  { 3600, B3600 },
#endif
#ifdef B4800
  { 4800, B4800 },
#endif
#ifdef B7200
  { 7200, B7200 },
#endif
#ifdef B9600
  { 9600, B9600 },
#endif
#ifdef EXTA
  { 19200, EXTA },
#endif
#ifdef EXTB
  { 38400, EXTB },
#endif
  { 0, 0 }
};

static int
setasyncmap(argcp, argvp)
int *argcp;
char ***argvp;
{
  unsigned long asyncmap;
  
  asyncmap = 0xffffffff;
  ++*argvp;
  sscanf(**argvp,"%lx",&asyncmap);
  ++*argvp;
  lcp_wantoptions[0].asyncmap = asyncmap;
  *argcp -= 2;
  return(1);
}

/*
 * setspeed - Set the speed.
 */
static int
setspeed(argcp, argvp)
int *argcp;
char ***argvp;
{
  int speed;
  struct speed *speedp;
  
  speed = atoi(**argvp);
  for (speedp = speeds; speedp->speed_int; speedp++)
    if (speed == speedp->speed_int) {
      inspeed = speedp->speed_val;
      --*argcp, ++*argvp;
      return (1);
    }
  return (0);
}


/*
 * setdevname - Set the device name.
 */
int setdevname(argcp, argvp)
int *argcp;
char ***argvp;
{
  char dev[DEVNAME_SIZE];
  char *cp = **argvp;
  struct stat statbuf;
  char *tty, *ttyname();
  
  if (strncmp("/dev/", cp, sizeof ("/dev/") - 1)) {
    (void) sprintf(dev, "/dev/%s", cp);
    cp = dev;
  }
  
  /*
   * Check if there is a device by this name.
   */
  if (stat(cp, &statbuf) < 0) {
    if (errno == ENOENT)
      return (0);
    syslog(LOG_ERR, cp);
    exit(1);
  }
  
  (void) strcpy(devname, cp);
  default_device = FALSE;
  --*argcp, ++*argvp;
  
  /*
   * If we haven't already decided to require authentication,
   * or we are running ppp on the control terminal, then we can
   * allow authentication to be requested.
   */
  if ((tty = ttyname(fileno(stdin))) == NULL)
    tty = ""; /* running from init means no stdin.  Null kills strcmp -KWK */
  if (lcp_wantoptions[0].neg_upap == 0 &&
      strcmp(devname, "/dev/tty") &&
      strcmp(devname, tty)) {
    lcp_wantoptions[0].neg_upap = 0;
    lcp_allowoptions[0].neg_upap = 1;
  }
  return (1);
}


/*
 * setipaddr - Set the IP address
 */
int
setipaddr(argcp, argvp)
int *argcp;
char ***argvp;
{
  u_long local, remote;
  struct hostent *hp;
  char *colon;
  
  /*
   * IP address pair separated by ":".
   */
  if ((colon = strchr(**argvp, ':')) == NULL)
    return (0);
  
  /*
   * If colon first character, then no local addr.
   */
  if (colon == **argvp) {
    local = 0l;
    ++colon;
  }
  else {
    *colon++ = '\0';
    if ((local = inet_addr(**argvp)) == -1) {
      if ((hp = gethostbyname(**argvp)) == NULL) {
	fprintf(stderr, "%s: Unknown host: %s", progname, **argvp);
	syslog(LOG_WARNING, "%s: Unknown host: %s", progname, **argvp);
	exit(1);
      }
      memcpy((char *) &local, hp->h_addr, hp->h_length);
    }
  }
  
  /*
   * If colon last character, then no remote addr.
   */
  if (*colon == '\0')
    remote = 0l;
  else {
    if ((remote = inet_addr(colon)) == -1) {
      if ((hp = gethostbyname(colon)) == NULL) {
	fprintf(stderr, "%s: Unknown host: %s", progname, colon);
	syslog(LOG_WARNING, "%s: Unknown host: %s", progname, colon);
	exit(1);
      }
      memcpy((char *) &remote, hp->h_addr, hp->h_length);
    }
  }
  
  ipcp_wantoptions[0].ouraddr = local;
  ipcp_wantoptions[0].hisaddr = remote;
  
 ret:
  --*argcp, ++*argvp;
  return (1);
}

static int
setnetmask(argcp, argvp)
int *argcp;
char ***argvp;
{
  u_long mask;
	
  --*argcp, ++*argvp;
  if (!strncasecmp (**argvp, "0x", 2))
    {                           /* netmask given in hex as 0xffffff00 */
      sscanf (**argvp + 2, "%lx", &mask);
      mask = htonl (mask);
    }
  else if ((mask = inet_addr (**argvp)) == -1) {
    fprintf(stderr, "%s: Invalid netmask %s\n", progname, **argvp);
    syslog(LOG_ERR,"Invalid netmask %s", **argvp);
    exit(1);
  }

  netmask.s_addr = mask;
  --*argcp, ++*argvp;
  return (1);
}

/*
 * setslip - Set slip framing mode.
 */
static int
setslip(argcp, argvp)
int *argcp;
char ***argvp;
{
  slip = TRUE;
  lcp_wantoptions[0].mru = SLIPMAX;
  if (argcp)
	--*argcp, ++*argvp;
  return (1);
}

/*
 * setsync - Set syncmode.
 */
static int
setsync(argcp, argvp)
int *argcp;
char ***argvp;
{
  syncmode = TRUE;
 
  lcp_wantoptions[0].neg_accompression = 0;
  lcp_allowoptions[0].neg_accompression = 0;
 
  lcp_wantoptions[0].neg_pcompression = 0;
  lcp_allowoptions[0].neg_pcompression = 0;
 
  lcp_wantoptions[0].neg_asyncmap = 0;
  lcp_allowoptions[0].neg_asyncmap = 0;
 
  --*argcp, ++*argvp;
  return (1);
}

/*
 * getuserpasswd - Get the user name and passwd.
 */
static void
getuserpasswd()
{
  upap[0].us_user = user;
  upap[0].us_userlen = strlen(upap[0].us_user);

  upap[0].us_passwd = passwd;
  upap[0].us_passwdlen = strlen(upap[0].us_passwd);
}

/*
 * getuseropt - Get the options from /etc/hosts.ppp for this user.
 */
int
getuseropt(user)
char *user;
{
  char buf[1024], *s;
  FILE *fp;
  int rc = 0;
  
  if ((fp = fopen(PPPHOSTS, "r")) == NULL)
    return (0);;
  
  /*
   * Loop till we find an entry for this user.
   */
  for (;;) {
    if (fgets(buf, sizeof (buf), fp) == NULL) {
      if (feof(fp))
	break;
      else {
	syslog(LOG_ERR, "fgets");
	exit(1);
      }
    }
    if ((s = strchr(buf, ' ')) == NULL)
      continue;
    *s++ = '\0';
    if (!strcmp(user, buf)) {
      rc = 1;
      break;
    }
  }
  fclose(fp);
  return (rc);
}

/*
 * Return user specified netmask. A value of zero means no netmask has
 * been set. 
 */
/* ARGSUSED */
u_long
GetMask(addr)
u_long addr;
{
  if(!netmask.s_addr) {
	if(IN_CLASSA(ntohl(addr)))
		netmask.s_addr = htonl(IN_CLASSA_NET);
	else if(IN_CLASSB(ntohl(addr)))
		netmask.s_addr = htonl(IN_CLASSB_NET);
	else if(IN_CLASSC(ntohl(addr)))
		netmask.s_addr = htonl(IN_CLASSC_NET);
  }
  return(netmask.s_addr);
}

/*
 * log_packet - format a packet and log it.
 */

char line[256];			/* line to be logged accumulated here */
char *linep;

void
log_packet(p, len, prefix)
u_char *p;
int len;
char *prefix;
{
    strcpy(line, prefix);
    linep = line + strlen(line);
    format_packet(p, len, pr_log, NULL);
    if (linep != line)
	syslog(LOG_DEBUG, "%s", line);
}

/*
 * format_packet - make a readable representation of a packet,
 * calling `printer(arg, format, ...)' to output it.
 */
void
format_packet(p, len, printer, arg)
u_char *p;
int len;
void (*printer) __ARGS((void *, char *, ...));
void *arg;
{
    int i, n;
    u_short proto;
    u_char x;

    if (len >= DLLHEADERLEN && p[0] == ALLSTATIONS && p[1] == UI) {
	p += 2;
	GETSHORT(proto, p);
	len -= DLLHEADERLEN;
	for (i = 0; i < N_PROTO; ++i)
	    if (proto == prottbl[i].protocol)
		break;
	if (i < N_PROTO) {
	    printer(arg, "[%s", prottbl[i].name);
	    n = (*prottbl[i].printpkt)(p, len, printer, arg);
	    printer(arg, "]");
	    p += n;
	    len -= n;
	} else {
	    printer(arg, "[proto=0x%x]", proto);
	}
    }

    for (; len > 0; --len) {
	GETCHAR(x, p);
	printer(arg, " %.2x", x);
    }
}

#ifdef __STDC__
#include <stdarg.h>

void
pr_log(void *arg, char *fmt, ...)
{
    int n;
    va_list pvar;
    char buf[256];

    va_start(pvar, fmt);
    vsprintf(buf, fmt, pvar);
    va_end(pvar);

    n = strlen(buf);
    if (linep + n + 1 > line + sizeof(line)) {
	syslog(LOG_DEBUG, "%s", line);
	linep = line;
    }
    strcpy(linep, buf);
    linep += n;
}

#else /* __STDC__ */
#include <varargs.h>

void
pr_log(arg, fmt, va_alist)
void *arg;
char *fmt;
va_dcl
{
    int n;
    va_list pvar;
    char buf[256];

    va_start(pvar);
    vsprintf(buf, fmt, pvar);
    va_end(pvar);

    n = strlen(buf);
    if (linep + n + 1 > line + sizeof(line)) {
	syslog(LOG_DEBUG, "%s", line);
	linep = line;
    }
    strcpy(linep, buf);
    linep += n;
}
#endif

/*
 * print_string - print a readable representation of a string using
 * printer.
 */
void
print_string(p, len, printer, arg)
char *p;
int len;
void (*printer) __ARGS((void *, char *, ...));
void *arg;
{
    int c;

    printer(arg, "\"");
    for (; len > 0; --len) {
	c = *p++;
	if (' ' <= c && c <= '~')
	    printer(arg, "%c", c);
	else
	    printer(arg, "\\%.3o", c);
    }
    printer(arg, "\"");
}

/*
 * novm - log an error message saying we ran out of memory, and die.
 */
void
novm(msg)
char *msg;
{
    syslog(LOG_ERR, "Virtual memory exhausted allocating %s\n", msg);
    quit();
}

/*
 * Read a word from a file.
 * Words are delimited by white-space or by quotes (").
 * Quotes, white-space and \ may be escaped with \.
 * \<newline> is ignored.
 */
int
getword(f, word, newlinep, filename)
FILE *f;
char *word;
int *newlinep;
char *filename;
{
    int c, len, escape;
    int quoted;

    *newlinep = 0;
    len = 0;
    escape = 0;
    quoted = 0;

    /*
     * First skip white-space and comments
     */
    while ((c = getc(f)) != EOF) {
	if (c == '\\') {
	    /*
	     * \<newline> is ignored; \ followed by anything else
	     * starts a word.
	     */
	    if ((c = getc(f)) == '\n')
		continue;
	    word[len++] = '\\';
	    escape = 1;
	    break;
	}
	if (c == '\n')
	    *newlinep = 1;	/* next word starts a line */
	else if (c == '#') {
	    /* comment - ignore until EOF or \n */
	    while ((c = getc(f)) != EOF && c != '\n')
		;
	    if (c == EOF)
		break;
	    *newlinep = 1;
	} else if (!isspace(c))
	    break;
    }

    /*
     * End of file or error - fail
     */
    if (c == EOF) {
	if (ferror(f)) {
	    perror(filename);
	    quit();
	}
	return 0;
    }

    for (;;) {
	/*
	 * Is this character escaped by \ ?
	 */
	if (escape) {
	    if (c == '\n')
		--len;			/* ignore \<newline> */
	    else if (c == '"' || isspace(c) || c == '\\')
		word[len-1] = c;	/* put special char in word */
	    else {
		if (len < MAXWORDLEN-1)
		    word[len] = c;
		++len;
	    }
	    escape = 0;
	} else if (c == '"') {
	    quoted = !quoted;
	} else if (!quoted && (isspace(c) || c == '#')) {
	    ungetc(c, f);
	    break;
	} else {
	    if (len < MAXWORDLEN-1)
		word[len] = c;
	    ++len;
	    if (c == '\\')
		escape = 1;
	}
	if ((c = getc(f)) == EOF)
	    break;
    }

    if (ferror(f)) {
	perror(filename);
	quit();
    }

    if (len >= MAXWORDLEN) {
	word[MAXWORDLEN-1] = 0;
	fprintf(stderr, "%s: warning: word in file %s too long (%.20s...)\n",
		progname, filename, word);
    } else
	word[len] = 0;

    return 1;
}

static int
setusehostname(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    usehostname = 1;
    return 1;
}

static int
setname(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    if (our_name[0] == 0) {
	strncpy(our_name, **argvp, MAXNAMELEN);
	our_name[MAXNAMELEN-1] = 0;
    }
    --*argcp, ++*argvp;
    return 1;
}

static int
setusername(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    strncpy(user, **argvp, MAXNAMELEN);
    user[MAXNAMELEN-1] = 0;
    --*argcp, ++*argvp;
    return 1;
}

static int
setremote(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    strncpy(remote_name, **argvp, MAXNAMELEN);
    remote_name[MAXNAMELEN-1] = 0;
    --*argcp, ++*argvp;
    return 1;
}

static int
setauth(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    auth_required = 1;
    return 1;
}

static int
setdologin(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    uselogin = 1;
    return 1;
}

static int
setpaptimeout(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    upap[0].us_timeouttime = atoi(**argvp);
    --*argcp, ++*argvp;
    return 1;
}

static int
setpapreqs(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    upap[0].us_maxtransmits = atoi(**argvp);
    --*argcp, ++*argvp;
    return 1;
}

static int
setchaptimeout(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    chap[0].timeouttime = atoi(**argvp);
    --*argcp, ++*argvp;
    return 1;
}

static int
setchapchal(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    chap[0].max_transmits = atoi(**argvp);
    --*argcp, ++*argvp;
    return 1;
}

static int
setchapintv(argcp, argvp)
int *argcp;
char ***argvp;
{
    --*argcp, ++*argvp;
    chap[0].chal_interval = atoi(**argvp);
    --*argcp, ++*argvp;
    return 1;
}
